import pygame, math
from pygame.locals import *

def dInput(string,default):
    """A version of the input() function with a default value if
        nothing is entered"""
    print(string," [",default,"]:",sep="",end=" ")
    inputValue = input()
    if inputValue == "":
        inputValue = default
    else:
        inputValue = inputValue
    return inputValue

# Define constants and initial conditions
h = .1
hConst = h/2    # Constant used in improved Euler
t = 0
tFinal = 100

r_x = r_y = r_z = 0
B_x = B_y = 0
B_z = float(dInput("Value for B_z",.3))

# Values for graphical output
width = height = 500
lineColor = (255,0,0)
bgColor = (255,255,255)
axisColor = (0,0,0)
xCenter = width/2
yCenter = height/2
rConst = math.pi/180     # Radians of rotation per frame

def xCoord(xValue):
    return xCenter + xValue*scale

def yCoord(yValue):
    return yCenter - yValue*scale

v_x = v_y = v_z = 1
x = "x"
y = "y" # Used for velocity and accel functions
z = "z"

rotationAxis = dInput("Rotate about which axis?",x)
rotationAngle = float(dInput("By how many degrees?",-30))


def velocity(component):
    if component == "x":
        return v_x
    if component == "y":
        return v_y
    if component == "z":
        return v_z

def acceleration(component,v_x,v_y,v_z):
    if component == "x":
        return v_y*B_z - v_z*B_y
    if component == "y":
        return v_z*B_x - v_x*B_z
    if component == "z":
        return v_x*B_y - v_y*B_x

def improvedEuler(r_x,r_y,r_z,v_x,v_y,v_z):
    m1x = velocity(x)
    m1y = velocity(y)
    m1z = velocity(z)
    k1x = acceleration(x,v_x,v_y,v_z)
    k1y = acceleration(y,v_x,v_y,v_z)
    k1z = acceleration(z,v_x,v_y,v_z)

    m2x = velocity(x) + h*k1x
    m2y = velocity(y) + h*k1y
    m2z = velocity(z) + h*k1z
    k2x = acceleration(x,m2x,m2y,m2z)
    k2y = acceleration(y,m2x,m2y,m2z)
    k2z = acceleration(z,m2x,m2y,m2z)

    r_x += hConst*(m1x+m2x)
    r_y += hConst*(m1y+m2y)
    r_z += hConst*(m1z+m2z)
    v_x += hConst*(k1x+k2x)
    v_y += hConst*(k1y+k2y)
    v_z += hConst*(k1z+k2z)

    return r_x,r_y,r_z,v_x,v_y,v_z

def xRotate(x,y,z,posOrNeg):
    """Rotate a point about the x-axis
        posOrNeg should be either 1 or -1"""
    angle = rConst*posOrNeg
    xPlot = x
    yPlot = y*math.cos(angle) - z*math.sin(angle)
    zPlot = z*math.cos(angle) + y*math.sin(angle)
    return xPlot,yPlot,zPlot

def yRotate(x,y,z,posOrNeg):
    """Rotate a point about the y-axis
        posOrNeg should be either 1 or -1"""
    angle = rConst*posOrNeg
    xPlot = math.cos(angle)*x - z*math.sin(angle)
    yPlot = y
    zPlot = math.sin(angle)*x + z*math.cos(angle)
    return xPlot,yPlot,zPlot

def zRotate(x,y,z,posOrNeg):
    """Rotate a point about the z-axis
        posOrNeg should be either 1 or -1"""
    angle = rConst*posOrNeg
    xPlot = math.cos(angle)*x - y*math.sin(angle)
    yPlot = math.sin(angle)*x + y*math.cos(angle)
    zPlot = z
    return xPlot,yPlot,zPlot

def rotatePoint(x,y,z,component,posOrNeg):
    if component == "x":
        return xRotate(x,y,z,posOrNeg)
    if component == "y":
        return yRotate(x,y,z,posOrNeg)
    if component == "z":
        return zRotate(x,y,z,posOrNeg)


# Autoscaling
r_xInitial,r_yInitial,r_zInitial = r_x,r_y,r_z
v_xInitial,v_yInitial,v_zInitial = v_x,v_y,v_z
tInitial = t
maxAbsValue = 0

while t < tFinal + hConst:
    r_x,r_y,r_z,v_x,v_y,v_z = improvedEuler(r_x,r_y,r_z,v_x,v_y,v_z)
    if abs(r_x) > maxAbsValue:
        maxAbsValue = abs(r_x)
    if abs(r_y) > maxAbsValue:
        maxAbsValue = abs(r_y)
    if abs(r_z) > maxAbsValue:
        maxAbsValue = abs(r_z)
    t += h
scale = width/(maxAbsValue)
r_x,r_y,r_z = r_xInitial,r_yInitial,r_zInitial
v_x,v_y,v_z = v_xInitial,v_yInitial,v_zInitial
t = tInitial

# Initialize screen, axes
pygame.init()
screen = pygame.display.set_mode((width,height))
screen.fill(bgColor)

pygame.draw.aaline(screen,axisColor,(0,yCenter),(width,yCenter),1)
pygame.draw.aaline(screen,axisColor,(xCenter,0),(xCenter,height),1)

def KE(v_x,v_y,v_z):
    return 0.5*(v_x**2 + v_y**2 + v_z**2)

KEInitial = KE(v_x,v_y,v_z)

while t < tFinal + hConst:
    r_xNew,r_yNew,r_zNew,v_xNew,v_yNew,v_zNew = improvedEuler(r_x,r_y,r_z,v_x,v_y,v_z)

    # Project 3D points rotated about an axis
    xPlot,yPlot,zPlot = rotatePoint(r_x,r_y,r_z,rotationAxis,rotationAngle)
    xPlotNew,yPlotNew,zPlotNew = rotatePoint(r_xNew,r_yNew,r_zNew,rotationAxis,rotationAngle)

    # Draw lines connecting projected points
    pygame.draw.aaline(screen,lineColor,(xCoord(xPlot),yCoord(yPlot)), \
        (xCoord(xPlotNew),yCoord(yPlotNew)),1)
    r_x,r_y,r_z,v_x,v_y,v_z = r_xNew,r_yNew,r_zNew,v_xNew,v_yNew,v_zNew
    t += h

pygame.display.flip()

KEFinal = KE(v_x,v_y,v_z)
KEError = (KEFinal - KEInitial)/KEInitial*100

print("KE conservation % error: ",KEError,"for step size",h)

running = True

while running:
    for event in pygame.event.get():
        # When you press the 'X' on the top right corner, the program closes
        if event.type == QUIT:
            running = False
        # When you press 'ESC' on the keyboard, the program closes
        if event.type == KEYDOWN and event.key == K_ESCAPE:
            running = False

pygame.display.quit()

# End of program
